# Third-party
import astropy.units as u
import numpy as np
import pytest

# Custom
from ..core import MockStream


def test_init():

    xyz = np.random.random(size=(3, 100)) * u.kpc
    vxyz = np.random.random(size=(3, 100)) * u.km / u.s
    t1 = np.random.random(size=100) * u.Myr

    lead_trail = np.empty(100, dtype='U1')
    lead_trail[::2] = 't'
    lead_trail[1::2] = 'l'

    stream = MockStream(xyz, vxyz)
    stream = MockStream(xyz, vxyz, release_time=t1)
    stream = MockStream(xyz, vxyz, lead_trail=lead_trail)  # noqa

    with pytest.raises(ValueError):
        MockStream(xyz, vxyz, release_time=t1[:-1])

    with pytest.raises(ValueError):
        MockStream(xyz, vxyz, lead_trail=lead_trail[:-1])


def test_one_burst():
    # Regression test: Tests a bug found by Helmer when putting all particles at
    # one timestep
    import gala.potential as gp
    import gala.dynamics as gd
    from gala.dynamics import mockstream as ms
    from gala.units import galactic

    # NFW MW with v_c = 232.8 km/s @ r = 8.2 kpc
    pot = gp.NFWPotential.from_circular_velocity(v_c=232.8*u.km/u.s,
                                                 r_s=15*u.kpc,
                                                 r_ref=8.2*u.kpc,
                                                 units=galactic)

    H = gp.Hamiltonian(pot)

    prog_w0 = gd.PhaseSpacePosition(pos=[10, 0, 0.] * u.kpc,
                                    vel=[0, 10, 0.] * u.km/u.s)

    dt = 1*u.Myr
    nsteps = 100
    orbit = H.integrate_orbit(prog_w0, dt=dt, n_steps=nsteps)

    r = orbit.spherical.distance

    n_array = np.zeros(orbit.t.size, dtype=int)
    argmin = r[0:150].argmin()
    n_array[argmin] = 1000

    df = ms.FardalStreamDF()

    dt = 1*u.Myr
    prog_mass = 2.5E4 * u.Msun
    prog_pot = gp.PlummerPotential(m=prog_mass, b=4*u.pc, units=galactic)

    gen = ms.MockStreamGenerator(df, H, progenitor_potential=prog_pot)

    stream, prog = gen.run(
        prog_w0, prog_mass,
        n_particles=n_array,
        dt=dt,
        n_steps=nsteps, progress=False)



""" Test dynamics plotting functions """

# Third-party
import numpy as np
import astropy.units as u
import pytest

# Project
from ..core import PhaseSpacePosition
from ..orbit import Orbit
from ..plot import plot_projections


def pytest_generate_tests(metafunc):
    if "obj" not in metafunc.fixturenames:
        return

    object_list = []

    norbits = 16
    object_list.append(
        PhaseSpacePosition(pos=np.random.random(size=3),
                           vel=np.random.random(size=3)))
    object_list.append(
        PhaseSpacePosition(pos=np.random.random(size=(3, norbits)),
                           vel=np.random.random(size=(3, norbits))))
    object_list.append(
        PhaseSpacePosition(pos=np.random.random(size=(3, norbits))*u.kpc,
                           vel=np.random.random(size=(3, norbits))*u.km/u.s))

    nsteps = 16
    object_list.append(Orbit(pos=np.random.random(size=(3, nsteps)),
                             vel=np.random.random(size=(3, nsteps)),
                             t=np.linspace(0, 1, nsteps)))
    object_list.append(Orbit(pos=np.random.random(size=(3, nsteps, 2)),
                             vel=np.random.random(size=(3, nsteps, 2)),
                             t=np.linspace(0, 1, nsteps)))
    object_list.append(Orbit(pos=np.random.random(size=(3, nsteps))*u.kpc,
                             vel=np.random.random(size=(3, nsteps))*u.km/u.s,
                             t=np.linspace(0, 1, nsteps)*u.Myr))

    # 2D
    object_list.append(
        PhaseSpacePosition(pos=np.random.random(size=(2, norbits)),
                           vel=np.random.random(size=(2, norbits))))
    object_list.append(Orbit(pos=np.random.random(size=(2, nsteps)),
                             vel=np.random.random(size=(2, nsteps)),
                             t=np.linspace(0, 1, nsteps)))

    test_names = [f'{obj.__class__.__name__}{i}'
                  for i, obj in enumerate(object_list)]

    metafunc.parametrize(['i', 'obj'],
                         list(enumerate(object_list)),
                         ids=test_names)


def test_plot_projections(i, obj):
    import matplotlib.pyplot as plt

    # Try executing the method - unfortunately no test of the actual figure
    # drawn!
    obj.plot()

    # Try with just 2D projection, and passing in a bunch of inputs...
    x = obj.xyz.value
    fig, axes = plt.subplots(1, 2)
    fig = plot_projections(x[:2], autolim=True, axes=axes,  # noqa
                           subplots_kwargs=dict(sharex=True),
                           labels=['x', 'y'],
                           plot_function=plt.plot,
                           marker='o', linestyle='--', color='r')


def test_animate(tmpdir, i, obj):
    if not isinstance(obj, Orbit):
        pytest.skip()

    # Try executing the method - unfortunately no test of the actual figure
    # drawn!
    fig, anim = obj.animate(segment_nsteps=3)
    anim.save(tmpdir / f'anim{i}.mp4')

    # test hiding the timestep label
    fig, anim = obj.animate(segment_nsteps=3, show_time=False)
    anim.save(tmpdir / f'anim{i}_no_time.mp4')

    if obj.ndim == 3:
        # Also try cylindrical, and sub-selecting components:
        fig, anim = obj.cylindrical.animate(components=['rho', 'z'])
        anim.save(tmpdir / f'anim{i}_cyl.mp4')



# Third-party
import numpy as np

__all__ = ["cartesian_to_poincare_polar"]


def cartesian_to_poincare_polar(w):
    r"""
    Convert an array of 6D Cartesian positions to Poincaré
    symplectic polar coordinates. These are similar to cylindrical
    coordinates.

    Parameters
    ----------
    w : array_like
        Input array of 6D Cartesian phase-space positions. Should have
        shape ``(norbits, 6)``.

    Returns
    -------
    new_w : :class:`~numpy.ndarray`
        Points represented in 6D Poincaré polar coordinates.

    """

    R = np.sqrt(w[..., 0] ** 2 + w[..., 1] ** 2)
    # phi = np.arctan2(w[..., 1], w[..., 0])
    phi = np.arctan2(w[..., 0], w[..., 1])

    vR = (w[..., 0] * w[..., 0 + 3] + w[..., 1] * w[..., 1 + 3]) / R
    vPhi = w[..., 0] * w[..., 1 + 3] - w[..., 1] * w[..., 0 + 3]

    # pg. 437, Papaphillipou & Laskar (1996)
    sqrt_2THETA = np.sqrt(np.abs(2 * vPhi))
    pp_phi = sqrt_2THETA * np.cos(phi)
    pp_phidot = sqrt_2THETA * np.sin(phi)

    z = w[..., 2]
    zdot = w[..., 2 + 3]

    new_w = np.vstack((R.T, pp_phi.T, z.T, vR.T, pp_phidot.T, zdot.T)).T
    return new_w
